/*
 * BrzozowskiDerivationRRG.cpp
 *
 *  Created on: 6. 3. 2014
 *      Author: tomas
 */

#include "BrzozowskiDerivationRRG.h"

using namespace alib;
using namespace alphabet;
using namespace grammar;
using namespace regexp;

namespace conversions
{

BrzozowskiDerivationRRG::BrzozowskiDerivationRRG( const RegExp & regexp ) : AbstractREtoRRGConverter( regexp )
{

}

BrzozowskiDerivationRRG::~BrzozowskiDerivationRRG( void )
{

}

RightRegularGrammar BrzozowskiDerivationRRG::convert( void )
{
    RegExpOptimize opt;

    // 1.
    RegExp V = opt.optimize( m_re );

    set<Symbol> alphabet = m_re.getAlphabet( );

    set<RegExp> N = { V };
    deque<set<RegExp>> Ni;

    Ni.push_back( set<RegExp>( ) );
    Ni.at( 0 ).insert( V );

    int i = 1;

    // 2.
    while( ! Ni.at( i - 1 ).empty( ) )
    {
        Ni.push_back( set<RegExp>( ) ); // initialize set Q_i

        for( const auto & regexp : Ni.at( i - 1 ) )
        {
            RegExpDerivation deriv( regexp );

            for( const auto & a : alphabet )
            {
                RegExpElement* dSymbol = new RegExpSymbol( a.getSymbol( ) );
                RegExp derived = deriv.derivation( list<RegExpElement*>( 1, dSymbol ) );
                derived = opt.optimize( derived );

                if( ! derived.isEmpty( ) )
                {
                    if( ! isInSet( derived, N ) ) // if this state has already been found, do not add
                        Ni.at( i ).insert( derived );

                    m_transitions.insert( Transition( regexp, Symbol( a.getSymbol( ) ), derived ) );
                }

                delete dSymbol;
            }
        }

        N.insert( Ni.at( i ).begin( ), Ni.at( i ).end( ) );
        i += 1;
    }

    // 3.

    for( const auto & s : alphabet )
        m_grammar.addTerminalSymbol( s.getSymbol( ) );

    NonTerminalBuilder builder( N, m_grammar );

    for( const auto & t : m_transitions )
    {
        list<Symbol> leftSide = { builder.getNonTerminal( t.m_from ) };
        list<Symbol> rightSide = { t.m_symbol, builder.getNonTerminal( t.m_to ) };
        Rule r( leftSide, rightSide );
        m_grammar.addRule( r );

        if( t.m_to.containsEmptyString( ) )
        {
            list<Symbol> rightSide = { t.m_symbol };
            Rule r( leftSide, rightSide );
            m_grammar.addRule( r );
        }
    }

    m_grammar.setStartSymbol( builder.getNonTerminal( V ) );

    if( V.containsEmptyString( ) )
    {
        list<Symbol> leftSide = { builder.getNonTerminal( V ) };
        list<Symbol> rightSide;

        if( m_grammar.isNonTerminalOnRightSideOfAnyRule( m_grammar.getStartSymbol( ) ) )
        {
            Symbol newStart = m_grammar.createUniqueNonTerminalSymbol( m_grammar.getStartSymbol( ).getSymbol( ), false );

            list<Symbol> leftSideNewStart = { newStart };
            for( const auto & rule : m_grammar.getRules( ) )
                if( rule.getLeftSide( ).front( ) == m_grammar.getStartSymbol( ) )
                    m_grammar.addRule( Rule( leftSideNewStart, rule.getRightSide( ) ) );

            m_grammar.setStartSymbol( newStart );
            m_grammar.addRule( Rule( leftSideNewStart, rightSide ) );
        }
        else
        {
            m_grammar.addRule( Rule ( leftSide, rightSide ) );
        }
    }

    return m_grammar;
}

// ----------------------------------------------------------------------------

BrzozowskiDerivationRRG::Transition::Transition( const RegExp & from, const Symbol & symb, const RegExp & to )
    : m_from( from ), m_to( to ), m_symbol( symb )
{

}

bool BrzozowskiDerivationRRG::Transition::operator<( const Transition & x ) const
{
    if( m_from != x.m_from )
        return m_from < x.m_from;
    else if( m_symbol != x.m_symbol )
        return m_symbol < x.m_symbol;
    else
        return m_to < x.m_to;
}

// ----------------------------------------------------------------------------

BrzozowskiDerivationRRG::NonTerminalBuilder::NonTerminalBuilder( const set<RegExp> & Q, RightRegularGrammar & grammar ) : m_grammar( grammar )
{
    m_nonTerminalId = 0;

    for( const auto & regexp : Q )
    {
        const string nonTerminalName = createNewName( );
        Symbol NT = m_grammar.createUniqueNonTerminalSymbol( nonTerminalName, false );

        m_nonTerminals.insert( pair<RegExp, Symbol>( regexp, NT ) );
    }
}

const Symbol & BrzozowskiDerivationRRG::NonTerminalBuilder::getNonTerminal( const RegExp & re ) const
{
    auto state = m_nonTerminals.find( re );
    if( state != m_nonTerminals.end( ) ) return state->second;

    throw AlibException( "BrzozowskiDerivationRRG::NonTerminalBuilder - Regular expression not found!" );
}

string BrzozowskiDerivationRRG::NonTerminalBuilder::createNewName( void )
{
    // http://en.wikipedia.org/wiki/Hexavigesimal

    unsigned int n = m_nonTerminalId ++;
    string name;
    do
    {
        unsigned int remainder = n % 26;
        name += ( char )( remainder + 'A' );
        n = (n - remainder) / 26;
    } while (n > 0);

    return string( name.rbegin( ), name.rend( ) );
}

} /* namespace conversions */
