/*
 * GlushkovTraversal.h
 *
 *  Created on: 13. 3. 2014
 *      Author: tomas
 */

#ifndef GLUSHKOVTRAVERSAL_H_
#define GLUSHKOVTRAVERSAL_H_

#include <set>
#include <list>

#include <regexp/RegExp.h>
#include <regexp/RegExpElement.h>
#include <regexp/Alternation.h>
#include <regexp/Concatenation.h>
#include <regexp/Iteration.h>
#include <regexp/RegExpSymbol.h>
#include <regexp/RegExpEmpty.h>
#include <regexp/RegExpEpsilon.h>

#include <AlibException.h>

namespace conversions
{

/**
 * RegExp tree traversal utils for Glushkov algorithm.
 */
class GlushkovTraversal
{
public:
    /**
     * Actually, this is just std::pair
     */
    struct Neighbours
    {
        const regexp::RegExpSymbol * m_first, * m_second;

        Neighbours( const regexp::RegExpSymbol * first, const regexp::RegExpSymbol * second );
        bool operator<( const Neighbours & x ) const;
    };

    /**
     * Returns all RegExpSymbols which can be initiating.
     */
    static std::set<const regexp::RegExpSymbol*> getLeftmostSymbolsInTree( const regexp::RegExp & re );

    /**
     * Returns all RegExpSymbols that can be terminating.
     */
    static std::set<const regexp::RegExpSymbol*> getRightmostSymbolsInTree( const regexp::RegExp & re );

    /**
     * Returns all pairs of RegExpSymbols that can be adjacent.
     */
    static std::set<Neighbours> getNeighbours( const regexp::RegExp & re );

    /**
     * Returns pointers to RegExpSymbols in order as they occur in regexp.
     *
     * @param re RegExp to probe
     * @return symbols of regexp tree in order of they occurence in regexp.
     */
    static std::list<const regexp::RegExpSymbol*> getSymbolsListInOrder( const regexp::RegExp & re );

private:
    static std::set<const regexp::RegExpSymbol*> getLeftmostSymbolsInTree( const regexp::RegExpElement * node );
    static std::set<const regexp::RegExpSymbol*> getLeftmostSymbolsInTree( const regexp::Alternation * node );
    static std::set<const regexp::RegExpSymbol*> getLeftmostSymbolsInTree( const regexp::Concatenation * node );
    static std::set<const regexp::RegExpSymbol*> getLeftmostSymbolsInTree( const regexp::Iteration * node );
    static std::set<const regexp::RegExpSymbol*> getLeftmostSymbolsInTree( const regexp::RegExpSymbol * node );
    static std::set<const regexp::RegExpSymbol*> getLeftmostSymbolsInTree( const regexp::RegExpEmpty * node );
    static std::set<const regexp::RegExpSymbol*> getLeftmostSymbolsInTree( const regexp::RegExpEpsilon * node );

    static std::set<const regexp::RegExpSymbol*> getRightmostSymbolsInTree( const regexp::RegExpElement * node );
    static std::set<const regexp::RegExpSymbol*> getRightmostSymbolsInTree( const regexp::Alternation * node );
    static std::set<const regexp::RegExpSymbol*> getRightmostSymbolsInTree( const regexp::Concatenation * node );
    static std::set<const regexp::RegExpSymbol*> getRightmostSymbolsInTree( const regexp::Iteration * node );
    static std::set<const regexp::RegExpSymbol*> getRightmostSymbolsInTree( const regexp::RegExpSymbol * node );
    static std::set<const regexp::RegExpSymbol*> getRightmostSymbolsInTree( const regexp::RegExpEmpty * node );
    static std::set<const regexp::RegExpSymbol*> getRightmostSymbolsInTree( const regexp::RegExpEpsilon * node );

    static std::set<Neighbours> getNeighbours( const regexp::RegExpElement * node );
    static std::set<Neighbours> getNeighbours( const regexp::Alternation * node );
    static std::set<Neighbours> getNeighbours( const regexp::Concatenation * node );
    static std::set<Neighbours> getNeighbours( const regexp::Iteration * node );
    static std::set<Neighbours> getNeighbours( const regexp::RegExpSymbol * node );
    static std::set<Neighbours> getNeighbours( const regexp::RegExpEmpty * node );
    static std::set<Neighbours> getNeighbours( const regexp::RegExpEpsilon * node );

    /**
     * @see GlushkovTraversal::getSymbolsListInOrder()
     */
    static void getSymbolsListInOrder( const regexp::RegExpElement * node, std::list<const regexp::RegExpSymbol*> & alphabet );
};

} /* namespace conversions */

#endif /* GLUSHKOVTRAVERSAL_H_ */
