/*
 * GasTexConverter.cpp
 *
 *  Created on: Jan 1, 2014
 *      Author: Martin Zak
 */

#include "GasTexConverter.h"

#include <set>
#include <map>
#include <typeinfo>
#include "AlibException.h"

using namespace std;

void GasTexConverter::convert(const Automaton& a, ostream& out) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialStates().find(state) != a.getInitialStates().end()){
			initial = true;
		}
		if(a.getFinalStates().find(state) != a.getFinalStates().end()){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state.getName();
		out << ")(,){";
		out << state.getName();
		out << "}\n";
	}

	try {
		const FSM& fsm = dynamic_cast<const FSM&>(a);
		transitions(fsm, out);
		out << "\\end{center}\n";
		out << "\\end{picture}\n";
		return;
	} catch (const bad_cast& e) {
	}

	try {
		const PDA& pda = dynamic_cast<const PDA&>(a);
		transitions(pda, out);
		out << "\\end{center}\n";
		out << "\\end{picture}\n";
		return;
	} catch (const bad_cast& e) {
	}

	try {
		const TM& tm = dynamic_cast<const TM&>(a);
		transitions(tm, out);
		out << "\\end{center}\n";
		out << "\\end{picture}\n";
		return;
	} catch (const bad_cast& e) {
	}

	throw new alib::AlibException("Unknown automaton type!");
}


string GasTexConverter::getStackSymbols(const list<Symbol>& stackSymbols) {
	if (stackSymbols.size() == 0) {
		return "$\\varepsilon$";
	}

	string symbols = "";
	int i=0;
	for (auto& popSymbol : stackSymbols) {
		if(i++ !=0) {
			symbols +=" ";
		}
		symbols +=  popSymbol.getSymbol();
	}
	return symbols;
}

string GasTexConverter::checkEpsilon(const string& inputSymbol) {
	if (inputSymbol.compare("") == 0) {
		return "$\\varepsilon$";
	} else {
		return inputSymbol;
	}
}

void GasTexConverter::printTransitionMap(
		const map<pair<string, string>, string> transitionMap, ostream& out) {
	for (auto& transition : transitionMap) {
		string from = transition.first.first;
		string to = transition.first.second;
		string symbols = transition.second;

		if (from == to) {
			out << "\\drawloop(";
			out << from;
			out << "){";
			out << symbols;
			out << "}\n";

		} else {
			out << "\\drawedge(";
			out << from;
			out << ",";
			out << to;
			out << "){";
			out << symbols;
			out << "}\n";
		}
	}
}

void GasTexConverter::transitions(const FSM& fsm, ostream& out) {
	map<pair<string, string>, string> transitionMap;
	for (auto& transition : fsm.getTransitions()) {
		pair<string, string> key(transition.getFrom().getName(), transition.getTo().getName());
		auto mapIterator = transitionMap.find(key);

		string symbol = checkEpsilon(transition.getInput().getSymbol());

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(
					pair<pair<string, string>, string>(key, symbol));
		} else {
			mapIterator->second += "," + symbol;
		}
	}
	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const PDA& pda, ostream& out) {
	map<pair<string, string>, string> transitionMap;

	for (auto& transition : pda.getTransitions()) {

		pair<string, string> key(transition.getFrom().getName(), transition.getTo().getName());
		auto mapIterator = transitionMap.find(key);

		string symbol = checkEpsilon(transition.getInput().getSymbol());

		symbol += ",";


		symbol += getStackSymbols(transition.getPop());
		symbol += "/";
		symbol += getStackSymbols(transition.getPush());

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(pair<pair<string, string>, string>(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	printTransitionMap(transitionMap, out);

}

void GasTexConverter::transitions(const TM& tm, ostream& out) {
	map<pair<string, string>, string> transitionMap;

	for (auto& transition : tm.getTransitions()) {

		pair<string, string> key(transition.getFrom().getName(), transition.getTo().getName());
		auto mapIterator = transitionMap.find(key);

		string symbol = checkEpsilon(transition.getInput().getSymbol());
		symbol += "/";
		symbol += checkEpsilon(transition.getOutput().getSymbol());
		symbol += ",";
		symbol += (std::string[] ) { "$\\leftarrow$", "$\\rightarrow$", "$\\times$" } [transition.getShift()];

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(pair<pair<string, string>, string>(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	printTransitionMap(transitionMap, out);

}

