#include <iostream>
#include <getopt.h>
#include <string.h>
#include <ctype.h>

#include "automaton/UnknownAutomaton.h"
#include "automaton/Automaton.h"
#include "automaton/FSM/FSM.h"
#include "automaton/PDA/PDA.h"
#include "AutomatonFactory.h"
#include "AlibException.h"

#include "fsm/FsmDeterminizer.h"
#include "idpda/IdpdaDeterminizer.h"
#include "vpa/VpaDeterminizer.h"
#include "vpa/VpaDeterminizer2.h"
#include "vpa/VpaDeterminizer3.h"
#include "rhdpda/RhdpdaDeterminizer.h"
#include "rhdpda/RhdpdaDeterminizer2.h"

using namespace std;
using namespace automaton;
using namespace determinization;
using namespace alib;

#define VERSION "0.0.1"


void printHelp() {
    cout << "adeterminize " << VERSION << endl;
    cout << "Determinize various types of automaton." << endl;
    cout << "Usage: adeterminize -t TYPE [SWITCH...]" << endl;
    cout << "Possible arguments:" << endl;
    cout << "  -t, --type=TYPE \t Specifies type of input automaton, possible values are 'FSM' for final-state machine, 'IDPDA' for input-driven pushdown automaton, 'VPA' for visible pushdown automaton and 'RHDPDA' for real-time height deterministic pushdown automaton." << endl;
    cout << "  -a, --algorithm=VERSION_OF_ALGORITHM \t Specifies version of algorithm. This argument works only with VPA or RHDPDA type. VPA determinization has 3 versions (use numbers 1, 2 and 3) and RHDPDA determinization has 2 versions (use numbers 1, 2). Default value is always number 1." << endl;
    cout << "  -h, --help \t Displays this help message." << endl;
    cout << endl;
}


Determinizer* getVpaDeterminizer(Automaton* automaton, string version) {
    if (version == "3") {
        return new vpa::VpaDeterminizer3((PDA*) automaton);

    } else if (version == "2") {
        return new vpa::VpaDeterminizer2((PDA*) automaton);
    }

    return new vpa::VpaDeterminizer((PDA*) automaton);
}


Determinizer* getRhdpdaDeterminizer(Automaton* automaton, string version) {
    if (version == "2") {
        return new rhdpda::RhdpdaDeterminizer2((PDA*) automaton);
    }

    return new rhdpda::RhdpdaDeterminizer((PDA*) automaton);
}


Determinizer* getDeterminizer(Automaton* automaton, string type, string version) {
    string _type;
    for (const auto& c : type) {
        _type.append(1, tolower(c));
    }
    type = _type;

    if (type == "fsm") {
        return new fsm::FsmDeterminizer((FSM*) automaton);

    } else if (type == "idpda") {
        return new idpda::IdpdaDeterminizer((PDA*) automaton);

    } else if (type == "vpa") {
        return getVpaDeterminizer(automaton, version);

    } else if (type == "rhdpda") {
        return getRhdpdaDeterminizer(automaton, version);
    }

    throw AlibException("Unknown type of input automaton. See help.");
}


int main(int argc, char** argv) {
    static struct option longOptions[] = {
      {"type", required_argument, NULL, 't'},
      {"algorithm", required_argument, NULL, 'a'},
      {"version", no_argument, NULL, 'v'},
      {"help", no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };

    int longIndex = 0;
    int opt = 0;
    string type;
    string version = "1";

    while ((opt = getopt_long(argc, argv, "t:a:vh", longOptions, &longIndex)) != -1) {
        switch(opt) {
            case 't':
                type.assign(optarg, strlen(optarg));
                break;
            case 'a':
                version.assign(optarg, strlen(optarg));
                break;
            case 'v':
            case 'h':
            default:
                printHelp();
                return 0;
        }
    }
    
    try {
        string input(istreambuf_iterator<char>(cin), (istreambuf_iterator<char>()));
        UnknownAutomaton unknownAutomaton = AutomatonFactory::fromString(input);
        Automaton* automaton = AutomatonFactory::buildAutomaton(unknownAutomaton);
        Determinizer* determinizer = getDeterminizer(automaton, type, version);
        Automaton* deterministicAutomaton = determinizer->determinize();
        deterministicAutomaton->toXML(cout);
        
        delete automaton;
        delete deterministicAutomaton;
        delete determinizer;
        
    } catch (AlibException& e) {
        cout << e.what() << endl;
    }

    return 0;
}
