#ifndef VPADETERMINIZER2_H_
#define VPADETERMINIZER2_H_

#include <string>
#include <set>
#include <map>
#include <vector>

#include "automaton/PDA/PDA.h"
#include "automaton/Transition.h"
#include "alphabet/Symbol.h"

#include "../common/Utils.h"
#include "../Determinizer.h"
#include "VpaUtils.h"
#include "VpaStructs.h"

using namespace std;
using namespace automaton;
using namespace alphabet;
using namespace determinization;

namespace determinization {
namespace vpa {

/**
 * Class for running second version of determinization algorithm on vpa.
 */
class VpaDeterminizer2 : public Determinizer
{

    private:

        /** Nondeterministic visibly pushdown automaton */
        PDA* nvpa;

        /** Deterministic visibly pushdown automaton */
        PDA* dvpa;

        /** Map of states of deterministic vpa, where key is the name of state and value is data about this state */
        map<string, StateData> states;

        /** Set of internal input symbols from nondeterministic automaton */
        set<Symbol> internalSymbols;

        /** Set of call input symbols from nondeterministic automaton */
        set<Symbol> callSymbols;

        /** Set of return input symbols from nondeterministic automaton */
        set<Symbol> returnSymbols;

        /** Vector with all possible S components of deterministic vpa */
        vector<SComponent> allSComponents;

        /** Vector with all possible R components of deterministic vpa */
        vector<RComponent> allRComponents;

        /**
         * Runs some initializiation stuff before determinization algorithm.
         */
        void initDeterminization();

        /**
         * Returns existing state from states map, if there is one with same name, or creates new one and adds it into
         *     states map and deterministic vpa.
         *
         * @param s S component of state
         * @param r R component of state
         * @return state of deterministic vpa
         */
        const State& getOrCreateState(const SComponent& s, const RComponent& r);

        /**
         * Returns S component that contains pairs where states are same.
         *
         * @param states set of states from which to create pairs
         * @return pairs where states are same
         */
        const SComponent getSComponentWithStatesIdentity(const set<State>& states);

        /**
         * Returns name of stack symbol of deterministic vpa which is created from given S and R component and input symbol.
         *     This method also ensures that this stack symbol is already added into deterministic vpa.
         *
         * @param s S component of stack symbol
         * @param r R component of stack symbol
         * @param input input symbol
         * @return name of stack symbol
         */
        string buildStackSymbolName(const SComponent& s, const RComponent& r, const Symbol& input);

    public:

        /**
         * @param nvpa nondeterministic visibly pushdown automaton given for determinization
         */
        VpaDeterminizer2(PDA* nvpa);

        /**
         * Runs determinization algorithm on nondeterministic vpa given in constructor.
         *
         * @return deterministic visibly pushdown automaton
         */
        Automaton* determinize();

};

}
}
#endif
