#include "VpaDeterminizer3.h"

namespace determinization {
namespace vpa {


string VpaDeterminizer3::buildStateName(const SComponent& s)
{
    string name = "{";
    for (const auto& statesPair : s.statesPairs) {
        if (&statesPair != &(*begin(s.statesPairs))) {
            name += ", ";
        }
        name += "('" + statesPair.state1.getName() + "', '" + statesPair.state2.getName() + "')";
    }
    name += "}";
    return name;
}


string VpaDeterminizer3::buildStackSymbolName(const SComponent& s, const Symbol& input)
{
    string name = "(" + this->buildStateName(s) + ")";
    name += ", " + input.getSymbol();
    try {
        this->dvpa->addStackSymbol(name);
    } catch (AutomatonException e) {
        /* Stack symbol already exists */
    }
    return name;
}


const State& VpaDeterminizer3::getOrCreateState(const SComponent& s)
{
    string stateName = this->buildStateName(s);

    map<string, StateData>::iterator statesIter = this->states.find(stateName);
    if (statesIter != this->states.end()) {
        return statesIter->second.state;
    }
    
    State state(stateName);
    StateData stateData(state, s);
    StateData& insertedData = this->states.insert(pair<string, StateData>(stateName, stateData)).first->second;
    this->dvpa->addState(state);
    return insertedData.state;
}


const SComponent VpaDeterminizer3::getSComponentWithStatesIdentity(const set<State>& states)
{
    return VpaUtils::getSComponentWithStatesIdentity(states);
}


void VpaDeterminizer3::initDeterminization()
{
    this->states.clear();
    this->internalSymbols.clear();
    this->callSymbols.clear();
    this->returnSymbols.clear();
    Utils::copyInputAlphabet(*this->nvpa, *this->dvpa);
    VpaUtils::divideInputAlphabet(*this->nvpa, this->internalSymbols, this->callSymbols, this->returnSymbols);
    this->allSComponents = VpaUtils::generateAllSComponents(this->nvpa->getStates());
    this->dvpa->addStackSymbol(VpaUtils::BOTTOM_OF_STACK_SYMBOL);
}


VpaDeterminizer3::VpaDeterminizer3(PDA* nvpa)
{
    this->nvpa = nvpa;
}


Automaton* VpaDeterminizer3::determinize()
{
    this->dvpa = new PDA();
    this->initDeterminization();
    
    const SComponent& initialS = this->getSComponentWithStatesIdentity(this->nvpa->getInitialStates());
    const State& initialState = this->getOrCreateState(initialS);
    this->dvpa->addInitialState(initialState);
    
    while (true) {
        StateData* unmarkedStateData = VpaUtils::getUnmarkedState(states);
        if (unmarkedStateData == NULL) {
            break;
        }
        
        // Internal symbols
        for (const auto& internalSymbol : internalSymbols) {
            SComponent sPrimed;
            SComponent& s = unmarkedStateData->s;
            for (const auto& statesPair : s.statesPairs) {
                const State& q = statesPair.state1;
                const State& qDoublePrimed = statesPair.state2;
                for (const auto& transition : this->nvpa->getTransitions()) {
                    if (transition.getFrom() == qDoublePrimed && transition.getInput() == internalSymbol) {
                        const State& qPrimed = transition.getTo();
                        sPrimed.statesPairs.insert(StatesPair(q, qPrimed));
                    }
                }
            }
            if (sPrimed.statesPairs.size() > 0) {
                const State& fromState = unmarkedStateData->state;
                const State& toState = this->getOrCreateState(sPrimed);
                const TransitionPDA transition(fromState, internalSymbol, toState);
                this->dvpa->addTransition(transition);
            }
        }
        
        // Call symbols
        for (const auto& callSymbol : callSymbols) {
            RComponent rPrimed;
            SComponent& s = unmarkedStateData->s;
            for (const auto& statesPair : s.statesPairs) {
                const State& q = statesPair.state2;
                for (const auto& transition : this->nvpa->getTransitions()) {
                    if (transition.getFrom() == q && transition.getInput() == callSymbol) {
                        const State& qPrimed = transition.getTo();
                        rPrimed.states.insert(qPrimed);
                    }
                }
            }
            if (rPrimed.states.size() > 0) {
                const SComponent& sPrimed = this->getSComponentWithStatesIdentity(rPrimed.states);
                const State& fromState = unmarkedStateData->state;
                const State& toState = this->getOrCreateState(sPrimed);
                string stackSymbolName = this->buildStackSymbolName(s, callSymbol);
                list<Symbol> pop = {};
                list<Symbol> push = {Symbol(stackSymbolName)};
                const TransitionPDA transition(fromState, callSymbol, toState, pop, push);
                this->dvpa->addTransition(transition);
            }
        }
        
        // Return symbols
        for (const auto& returnSymbol : returnSymbols) {
            SComponent& s = unmarkedStateData->s;
            
            // Empty stack
            list<Symbol> bottomOfStack = {VpaUtils::BOTTOM_OF_STACK_SYMBOL};
            SComponent sPrimed;
            for (const auto& statesPair : s.statesPairs) {
                const State& q = statesPair.state1;
                const State& qDoublePrimed = statesPair.state2;
                for (const auto& transition : this->nvpa->getTransitions()) {
                    if (transition.getFrom() == qDoublePrimed && transition.getInput() == returnSymbol &&
                            transition.getPop() == bottomOfStack) {
                        const State& qPrimed = transition.getTo();
                        sPrimed.statesPairs.insert(StatesPair(q, qPrimed));
                    }
                }
            }
            if (sPrimed.statesPairs.size() > 0) {
                const State& fromState = unmarkedStateData->state;
                const State& toState = this->getOrCreateState(sPrimed);
                list<Symbol> pop = {VpaUtils::BOTTOM_OF_STACK_SYMBOL};
                list<Symbol> push = {};
                const TransitionPDA transition(fromState, returnSymbol, toState, pop, push);
                this->dvpa->addTransition(transition);
            }
            
            // Otherwise
            for (const auto& callSymbol : callSymbols) {
                set<StatesPair> update;
                for (const auto& sPair : s.statesPairs) {
                    const State& q1 = sPair.state1;
                    const State& q2 = sPair.state2;
                    for (const auto& transitionPush : this->nvpa->getTransitions()) {
                        for (const auto& transitionPop : this->nvpa->getTransitions()) {
                            const State& q = transitionPush.getFrom();
                            const State& qPrimed = transitionPop.getTo();
                            if (transitionPush.getInput() == callSymbol && transitionPush.getTo() == q1 &&
                                    transitionPop.getInput() == returnSymbol && transitionPop.getFrom() == q2 &&
                                    transitionPush.getPush() == transitionPop.getPop()) {
                                update.insert(StatesPair(q, qPrimed));
                            }
                        }
                    }
                }
                if (update.size() == 0) {
                    continue;
                }
                for (const auto& sPrimed : this->allSComponents) {
                    SComponent sDoublePrimed;
                    for (const auto& updateItem : update) {
                        const State& q3 = updateItem.state1;
                        const State& qPrimed = updateItem.state2;
                        for (const auto& sPrimedPair : sPrimed.statesPairs) {
                            if (sPrimedPair.state2 == q3) {
                                const State& q = sPrimedPair.state1;
                                sDoublePrimed.statesPairs.insert(StatesPair(q, qPrimed));
                            }
                        }
                    }
                    if (sDoublePrimed.statesPairs.size() > 0) {
                        const State& fromState = unmarkedStateData->state;
                        const State& toState = this->getOrCreateState(sDoublePrimed);
                        string stackSymbolName = this->buildStackSymbolName(sPrimed, callSymbol);
                        list<Symbol> pop = {Symbol(stackSymbolName)};
                        list<Symbol> push = {};
                        const TransitionPDA transition(fromState, returnSymbol, toState, pop, push);
                        this->dvpa->addTransition(transition);
                    }
                }
            }
        }
    }
    
    // Final states
    for (const auto& stateIter : states) {
        const StateData& stateData = stateIter.second;
        set<State> secondStates;
        for (const auto& sPair : stateData.s.statesPairs) {
            secondStates.insert(sPair.state2);
        }
        if (Utils::containFinalState(secondStates, *this->nvpa)) {
            this->dvpa->addFinalState(stateData.state);
        }
    }
    
    return this->dvpa;
}


}
}
