#include "VpaUtils.h"

namespace determinization {
namespace vpa {


const Symbol VpaUtils::BOTTOM_OF_STACK_SYMBOL("⊥");


string VpaUtils::buildStateName(const SComponent& s, const RComponent& r) {
    string name = "{";
    for (const auto& statesPair : s.statesPairs) {
        if (&statesPair != &(*begin(s.statesPairs))) {
            name += ", ";
        }
        name += "('" + statesPair.state1.getName() + "', '" + statesPair.state2.getName() + "')";
    }
    name += "}, {";
    for (const auto& state : r.states) {
        if (&state != &(*begin(r.states))) {
            name += ", ";
        }
        name += "'" + state.getName() + "'";
    }
    name += "}";
    return name;
}


const State& VpaUtils::getOrCreateState(const SComponent& s, const RComponent& r, map<string, StateData>& states, PDA& dvpa) {
    string stateName = buildStateName(s, r);

    map<string, StateData>::iterator statesIter = states.find(stateName);
    if (statesIter != states.end()) {
        return statesIter->second.state;
    }
    
    State state(stateName);
    StateData stateData(state, s, r);
    StateData& insertedData = states.insert(pair<string, StateData>(stateName, stateData)).first->second;
    dvpa.addState(state);
    return insertedData.state;
}


const SComponent VpaUtils::getSComponentWithStatesIdentity(const set<State>& states) {
    set<StatesPair> statesPairs;
    for (const auto& state : states) {
        statesPairs.insert(StatesPair(state, state));
    }
    return SComponent(statesPairs);
}


void VpaUtils::divideInputAlphabet(PDA& vpa, set<Symbol>& internalSymbols, set<Symbol>& callSymbols, set<Symbol>& returnSymbols) {
    for (const auto& input : vpa.getInputAlphabet()) {
        for (const auto& transition : vpa.getTransitions()) {
            if (transition.getInput() == input) {
                if (transition.getPush().size() != 0) {
                    callSymbols.insert(input);
                } else if (transition.getPop().size() != 0) {
                    returnSymbols.insert(input);
                } else {
                    internalSymbols.insert(input);
                }
                break;
            }
        }
    }
}


string VpaUtils::buildStackSymbolName(const SComponent& s, const RComponent& r, const Symbol& input, PDA& dvpa) {
    string name = "(" + buildStateName(s, r) + ")";
    name += ", " + input.getSymbol();
    try {
        dvpa.addStackSymbol(name);
    } catch (AutomatonException e) {
        /* Stack symbol already exists */
    }
    return name;
}


vector<SComponent> VpaUtils::generateAllSComponents(const set<State>& states)
{
    vector<StatesPair> possibleStatesPairs;
    for (const auto& state1 : states) {
        for (const auto& state2 : states) {
            possibleStatesPairs.push_back(StatesPair(state1, state2));
        }
    }
    vector<SComponent> possibleS;
    int possibleSSize = pow(2, possibleStatesPairs.size());
    for (int i = 0; i < possibleSSize; i++) {
        set<StatesPair> statesPairsSet;
        int mask = 1;
        for (const auto& statesPair : possibleStatesPairs) {
            if ((i & mask) != 0) {
                statesPairsSet.insert(statesPair);
            }
            mask <<= 1;
        }
        possibleS.push_back(SComponent(statesPairsSet));
    }
    return possibleS;
}


vector<RComponent> VpaUtils::generateAllRComponents(const set<State>& states)
{
    vector<RComponent> possibleR;
    int possibleRSize = pow(2, states.size());
    for (int i = 0; i < possibleRSize; i++) {
        set<State> statesSet;
        int mask = 1;
        for (const auto& state : states) {
            if ((i & mask) != 0) {
                statesSet.insert(state);
            }
            mask <<= 1;
        }
        possibleR.push_back(RComponent(statesSet));
    }
    return possibleR;
}


StateData* VpaUtils::getUnmarkedState(map<string, StateData>& states) {
    StateData* unmarkedStateData = NULL;
    for (auto& stateIter : states) {
        StateData& stateData = stateIter.second;
        if (!stateData.isMarked) {
            unmarkedStateData = &stateData;
            unmarkedStateData->isMarked = true;
            return unmarkedStateData;
        }
    }
    return NULL;
}


}
}
