/*
 * PDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: martin
 */

#include "PDA.h"
#include "../AutomatonPrinter.h"
#include "../exception/AutomatonException.h"
#include <algorithm>

namespace automaton {

using namespace std;

PDA::PDA() {

}

void PDA::removeState(const State& state) {
	for (set<State>::const_iterator initialState = initialStates.begin(); initialState != initialStates.end();
			initialState++) {
		if (*initialState == state) {
			throw AutomatonException("State \"" + state.getName() + "\" is initial state.");
		}
	}

	for (set<State>::const_iterator finalState = finalStates.begin(); finalState != finalStates.end(); finalState++) {
		if (*finalState == state) {
			throw AutomatonException("State \"" + state.getName() + "\" is final state.");
		}
	}

	for (set<TransitionPDA>::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		if (state == t->getFrom() || state == t->getTo())
			throw AutomatonException("State \"" + state.getName() + "\" is used in transition.");
	}

	int removed = states.erase(state);
	if (!removed)
		throw AutomatonException("State \"" + state.getName() + "\" doesn't exist.");
}

void PDA::removeInputSymbol(const Symbol& symbol) {
	for (set<TransitionPDA>::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		if (symbol == t->getInput())
			throw AutomatonException("Symbol \"" + symbol.getSymbol() + "\" is used in transition.");
	}

	int removed = inputAlphabet.erase(symbol);
	if (!removed)
		throw AutomatonException("Symbol \"" + symbol.getSymbol() + "\" doesn't exist.");
}

void PDA::addStackSymbol(const Symbol& symbol) {
	pair<set<Symbol>::iterator, bool> ret = stackAlphabet.insert(symbol);
	if (!ret.second)
		throw AutomatonException("Stack symbol \"" + symbol.getSymbol() + "\" already exists.");
}

void PDA::removeStackSymbol(const Symbol& symbol) {
	for (set<TransitionPDA>::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		for (list<Symbol>::const_iterator popSymbol = t->getPop().begin(); popSymbol != t->getPop().end();
				popSymbol++) {
			if (symbol == (*popSymbol))
				throw AutomatonException("Stack symbol \"" + symbol.getSymbol() + "\" is used in transition.");
		}
		for (list<Symbol>::const_iterator pushSymbol = t->getPush().begin(); pushSymbol != t->getPush().end();
				pushSymbol++) {
			if (symbol == (*pushSymbol))
				throw AutomatonException("Stack symbol \"" + symbol.getSymbol() + "\" is used in transition.");
		}
	}

	for(list<Symbol>::const_iterator startSymbol = startSymbols.begin(); startSymbol != startSymbols.end(); startSymbol++){
		if(*startSymbol == symbol) {
			throw AutomatonException("Stack symbol \"" + symbol.getSymbol() + "\" is start symbol.");
		}
	}

	int removed = stackAlphabet.erase(symbol);
	if (!removed)
		throw AutomatonException("Stack symbol \"" + symbol.getSymbol() + "\" doesn't exist.");

}

const set<Symbol>& PDA::getStackAlphabet() const {
	return stackAlphabet;
}

void PDA::addTransition(const TransitionPDA& transition) {
	if (states.find(transition.getFrom()) == states.end()) {
		throw AutomatonException("State \"" + transition.getFrom().getName() + "\" doesn't exist.");
	}

	if (transition.getInput().getSymbol().compare("") != 0) {
		if (inputAlphabet.find(transition.getInput()) == inputAlphabet.end()) {
			throw AutomatonException("Input symbol \"" + transition.getInput().getSymbol() + "\" doesn't exist.");
		}
	}

	if (states.find(transition.getTo()) == states.end()) {
		throw AutomatonException("State \"" + transition.getFrom().getName() + "\" doesn't exist.");
	}

	list<Symbol>::const_iterator popSymbol = transition.getPop().begin();
	while (popSymbol != transition.getPop().end()) {
		if (stackAlphabet.find(*popSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + popSymbol->getSymbol() + "\" doesn't exist.");
		}
		popSymbol++;
	}

	list<Symbol>::const_iterator pushSymbol = transition.getPush().begin();
	while (pushSymbol != transition.getPush().end()) {
		if (stackAlphabet.find(*pushSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + pushSymbol->getSymbol() + "\" doesn't exist.");
		}
		pushSymbol++;
	}

	pair<set<TransitionPDA>::iterator, bool> ret = transitions.insert(transition);
	if (!ret.second) {
		throw AutomatonException("Transition already exists.");
	}
}

void PDA::removeTransition(const TransitionPDA& transition) {
	int removed = transitions.erase(transition);
	if (!removed) {
		throw AutomatonException("Transition doesn't exist.");
	}
}

const set<TransitionPDA>& PDA::getTransitions() const {
	return transitions;
}

void PDA::setStartSymbols(const list<Symbol>& start) {
	for (list<Symbol>::const_iterator symbol = start.begin(); symbol != start.end(); symbol++) {
		if (stackAlphabet.find(*symbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + symbol->getSymbol() + "\" doesn't exist.");
		}
	}

	this->startSymbols = start;
}

const list<Symbol>& PDA::getStartSymbols() const {
	return startSymbols;
}

void PDA::toXML(ostream& out) const {
	AutomatonPrinter::toXML(*this, out);
}

} /* namespace automaton */
